
/***********************************************************************
 *                                                                     *
 *                               Encoder                               *
 *                                                                     *
 ***********************************************************************
  The encoded file will look like the following:
   - Header
       - Magic
       - Bitmaps Section total size
       - Bitmaps Section offset
       - Graymaps Section total size
       - Graymaps Section offset
       - Pages Section total size
       - Pages Section offset
       - Bookmarks Section total size
       - Bookmarks Section offset
   - Bitmaps
       - #bitmaps
       - bitmap table
          - offset *N
       - bitmap data
          [size[4] bitmap]
   - Graymaps
       - #graymaps
       - graymap table
          - offset * N
       - graymap data
          [size[4] graymap]
   - Pages
       - #pages
       - page table
          offset*N
       - page data
          size[4], decomp_command_size[4], width[2], height[2], background[1] commands..
           command: bit, gray, rect
             bit: 'B' X[2] Y[2] B[4]
             gray: 'G' X[2] Y[2] G[4]
             rect: 'R' X[2] Y[2] W[2] H[2] color[1]
    Bookmarks are described in bmk_parse.c. The top level structure is:
   - Bookmarks
      - size
      - data
 ***********************************************************************/
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <zlib.h>
#include "pngio.h"
#include "gmap.h"
#include "exops.h"
#include "bitmap.h"
#include "bmk.h"

#include "buffer.h"
#include "objmap.h"
#include "ptrbuf.h"



#include "util-encode-2.c"

static int Obyte(unsigned char *D, unsigned int V) 
  { D[0]= V; return 1; }
static int Oword(unsigned char *D, unsigned int V) 
  { D[0]= V; D[1]= V>>8; return 2; } 
static int Odoub(unsigned char *D, unsigned int V) 
  { D[0]= V; D[1]= V>>8; D[2]= V>>16; D[3]= V>>24; return 4; } 



/***********************************************************************
 *                                                                     *
 *                              Commands                               *
 *                                                                     *
 ***********************************************************************
  Initially, we collect commands in the following struct. This is
  because we want to output bitmaps and graymaps in a sorted order.
  After finishing the document, we will find out the sorted order for
  the images and put proper index numbers into the commands.
 ***********************************************************************/

typedef struct
{
  unsigned char *data;
  int len;
  int number;
} bentry_t;

typedef struct
{
  unsigned char *data;
  int len;
  int number;
} gentry_t;

typedef struct
{
  char type;
  uint16_t x,y;
  union
  {
    struct { bentry_t *be; } B;
    struct { gentry_t *ge; } G;
    struct { uint16_t w,h; uint8_t color; } R;
  } d;
} command_t;

typedef struct { int ncommands; command_t *commands;  int w, h; } page_t;

static int bitmap_count, graymap_count;
static objmap_t *bitmap_tree;
static objmap_t *graymap_tree;
static ptrbuf_t *page_buffer;
static int page_count;

static bentry_t **bitmap_array;
static gentry_t **graymap_array;
static page_t   **page_array;

static int bitmap_hits, graymap_hits;

static bentry_t* register_bitmap(unsigned char *bmap, int blen)
{
  bentry_t *be;

  be= objmap_find(bitmap_tree, bmap, blen);
  if (be)
  {
    bitmap_hits++;
    free(bmap);
    return be;
  }

  be= malloc(sizeof(*be));
  be->data= bmap;
  be->len= blen;
  objmap_insert(bitmap_tree,bmap,blen, be);
  bitmap_count++;
  return be;
}

static const unsigned char minimal_png[]= {
0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,
0x00,0x0d,0x49,0x48,0x44,0x52,0x00,0x00,0x00,0x01,
0x00,0x00,0x00,0x01,0x08,0x00,0x00,0x00,0x00,0x3a,
0x7e,0x9b,0x55,0x00,0x00,0x00,0x0a,0x49,0x44,0x41,
0x54,0x78,0x9c,0x63,0x62,0x00,0x00,0x00,0x06,0x00,
0x03,0x36,0x37,0x7c,0xa8,0x00,0x00,0x00,0x00,0x49,
0x45,0x4e,0x44,0xae,0x42,0x60,0x82, };




static gentry_t* register_graymap(gmap_t *src, pect_t *R)
{
  unsigned char *gmap;
  int glen;
  unsigned char *Pmap; unsigned int Plen;
  int w,h,x,y;
  gentry_t *ge;

  w= R->x1-R->x0;
  h= R->y1-R->y0;

  gmap= malloc(w*h);

  for(y=0;y<h;y++)
    for(x=0;x<w;x++)
      gmap[ x + y*w ]= src->data[ (y+R->y0)*src->w + (x+R->x0) ];

  if (encode_png_g8(w,h, gmap, &Plen, &Pmap))
  {
    Pmap= (unsigned char*) minimal_png;
    Plen= sizeof(minimal_png);
    w= h= 1;
  }
  free(gmap);

  glen= Plen + 4;
  gmap= malloc( glen );
  gmap[0]= w; gmap[1]= w>>8;
  gmap[2]= h; gmap[3]= h>>8;
  memcpy(gmap+4, Pmap, Plen);
  if (Pmap!=(unsigned char*) minimal_png) free(Pmap);

  ge= objmap_find(graymap_tree, gmap, glen);
  if (ge)
  {
    graymap_hits++;
    free(gmap);
    return ge;
  }

  ge= malloc(sizeof(*ge));
  ge->data= gmap;
  ge->len= glen;
  objmap_insert(graymap_tree,gmap,glen, ge);
  graymap_count++;
  return ge;
}

static void command_rect(buffer_t *CB,pect_t *R, int color,int *NC)
{
  command_t C;
  C.type= 'R';
  C.x= R->x0;
  C.y= R->y0;
  C.d.R.w= R->x1 - R->x0;
  C.d.R.h= R->y1 - R->y0;
  C.d.R.color= color;
  buffer_write(CB, &C, sizeof(C));
  *NC= *NC+1;
}

static void command_bitm(buffer_t *CB,pect_t *R, bentry_t *be,int *NC)
{
  command_t C;
  C.type= 'B';
  C.x= R->x0;
  C.y= R->y0;
  C.d.B.be= be;
  buffer_write(CB, &C, sizeof(C));
  *NC= *NC+1;
}

static void command_gray(buffer_t *CB,pect_t *R, gentry_t *ge,int *NC)
{
  command_t C;
  C.type= 'G';
  C.x= R->x0;
  C.y= R->y0;
  C.d.G.ge= ge;
  buffer_write(CB, &C, sizeof(C));
  *NC= *NC+1;
}

static int uniform_rectangle(gmap_t *src, pect_t *R)
{
  unsigned int P;
  int x,y;
  P= src->data[ R->x0+R->y0*src->w ];
  for(y=R->y0;y<R->y1;y++)
   for(x=R->x0;x<R->x1;x++)
     if (src->data[x+y*src->w]!=P) return 0;
  return 1;
}

static void extract_bitmap(buffer_t *CB,gmap_t *src, pect_t *R,int *NC)
{
  unsigned char *bmap; int blen;
  bentry_t *be;
  bmap= graymap_to_bitmap(src->data, src->w,src->h,
                          R->x0, R->y0, R->x1-R->x0, R->y1-R->y0,
                          0x00, &blen);
  be= register_bitmap(bmap, blen);
  command_bitm(CB,R, be,NC);
  gmap_clear(src, R, gmap_choose_bg(src,R,0xff,0x00));
}

static void copy_and_clear(gmap_t *dst, gmap_t *src, pect_t *R)
{
  int x,y;
  for(y=R->y0;y<R->y1;y++)
    for(x=R->x0;x<R->x1;x++)
    {
      dst->data[x+y*dst->w]= src->data[x+y*src->w];
      src->data[x+y*src->w]= 0xff;
    }
}

static int check_for_bitmap(gmap_t *src, int x,int y, unsigned char *other)
{
  unsigned char P;
  P= src->data[x+y*src->w];
  if (P==0x00) return 0;
  if (*other==0x00) { *other= P; return 0; }
  if (P!=*other) return 1;
            else return 0;
}

static int suitable_for_bitmap(gmap_t *src, pect_t *R)
{
  int x,y;
  unsigned char othercolor= 0x00;

  int sx,sy, ex,ey;
  sx= R->x0; 
  sy= R->y0;
  ey= R->y1;
  ex= R->x1; 

  if (ex-sx<2 || ey-sy<2) return 0;

  for(y=sy;y<ey;y++) for(x=sx;x<ex;x++)
     if (check_for_bitmap(src,x,y, &othercolor)) return 0;
  if (sx>0) for(y=sy;y<ey;y++) 
     if (check_for_bitmap(src,sx-1,y, &othercolor)) return 0;
  if (ex<src->w) for(y=sy;y<ey;y++) 
     if (check_for_bitmap(src,ex,y, &othercolor)) return 0;
  if (sy>0) for(x=sx;x<ex;x++)
     if (check_for_bitmap(src,x,sy-1, &othercolor)) return 0;
  if (ey<src->h) for(x=sx;x<ex;x++)
     if (check_for_bitmap(src,x,ey, &othercolor)) return 0;
  return 1;
}

static void reduce_colors(gmap_t *src, int col)
{
  static unsigned char rmap[256];
  static int rmap_ok;
  int i;

  if (!rmap_ok)
  {
    col= 256/col;
    for(i=0;i<256;i++)
    {
      int V;
      V= ( i + (col/2-1) )/col *col;
      if (V>255) V= 255;
      rmap[i]= V;
    }
    rmap_ok= 1;
  }
  int npix= src->w*src->h;
  for(i=0;i<npix;i++)
   src->data[i]= rmap[src->data[i]];
}

static void remove_lines(gmap_t *src, buffer_t *CB, int *NC)
{
  int threshold,SS;
  unsigned int col;
  pect_t R;
  threshold= 50;
  
  SS= 0;
  while(find_hor_line(src, &R, 0xff, threshold, &SS,&col))
  {
    command_rect(CB,&R, col, NC); 
    gmap_clear(src, &R, 0xff); // gmap_choose_bg(src,&R, 0xff,0x00));
  }
  SS= 0;
  while(find_ver_line(src, &R, 0xff, threshold, &SS,&col))
  {
    command_rect(CB,&R, col, NC);
    gmap_clear(src, &R, 0xff); // gmap_choose_bg(src,&R,0xff,0x00));
  }
}

static void remove_trivial_bitmaps
      (gmap_t *src,gmap_t *tmp, buffer_t *CB,int *NC)
{
  int SS;
  pect_t R;

  SS= 0;
  while(partition(src,&R,0xff,&SS))
  {
    if (suitable_for_bitmap(src,&R))
      extract_bitmap(CB,src, &R, NC);
    else
      copy_and_clear(tmp, src, &R);
  }
      // put back the things we had found before,
      // they will be part of images
  SS= 0;
  while(partition(tmp,&R,0xff,&SS))
      copy_and_clear(src, tmp, &R);
}

static void encode_images(gmap_t *src,buffer_t *CB,int *NC)
{
  pect_t R;
  int SS;

  SS= 0;
  while(partition(src,&R,0xff,&SS))
  {
    if (uniform_rectangle(src,&R))
    {
      command_rect(CB,&R, src->data[R.x0+R.y0*src->w],NC);
    }
    else
    {
      gentry_t *ge;
      ge= register_graymap(src, &R);
      command_gray(CB,&R, ge,NC);
    }
    gmap_clear(src, &R, 0xff);
  }
}

static int is_dual_color_with_black(gmap_t *src, pect_t *R, unsigned char *OC)
{
  int x,y;
  int have_black;
  unsigned char other, P;

  have_black= 0;
  other= 0x00;

  for(y=R->y0;y<R->y1;y++)
   for(x=R->x0;x<R->x1;x++)
   {
     P= src->data[x+y*src->w];
     if (P==0x00) { have_black=1; continue; }
     if (other==0x00) { other= P; continue; }
     if (P!=other) return 0;
   }
  if (!have_black) return 0;
  *OC= other;
  return 1;
}

static gmap_t *mkgtemp(gmap_t *src)
{
  gmap_t *tmp;
  tmp=malloc(sizeof(*tmp));
  tmp->data= malloc(src->w*src->h);
  tmp->w= src->w; 
  tmp->h= src->h;
  memset(tmp->data, 0xff, src->w*src->h);
  return tmp;
}
static void free_gmap(gmap_t *gm)
{
  free(gm->data);
  free(gm);
}

static void extract_bitmaps_with_background
         (gmap_t *src, gmap_t *tmp, gmap_t *tmp2, buffer_t *CB, int *NC)
{
  int SS, SK;
  pect_t R,U;
  unsigned char OC;
  int x,y;

  SS= 0;
  while(partition(src,&R,0xff,&SS))
  {
    if (!is_dual_color_with_black(src,&R,&OC))
    {
      copy_and_clear(tmp, src, &R);
      continue;
    }

    copy_and_clear(tmp2, src, &R);
    for(y=R.y0;y<R.y1;y++)
     for(x=R.x0;x<R.x1;x++)
       if (tmp2->data[x+y*tmp2->w]==OC)
          tmp2->data[x+y*tmp2->w]= 0xff;

    SK= 0;
    while(partition(tmp2, &U, 0xff, &SK))
      extract_bitmap(CB,tmp2, &U, NC);

    command_rect(CB, &R, OC, NC);
  }

  SS= 0;
  while(partition(tmp,&R,0xff,&SS))
      copy_and_clear(src, tmp, &R);
}

static command_t* process_image(gmap_t *src,int *Rncommands)
{
  pect_t R;
  buffer_t *command_buffer;
  int ncommands;
  command_t *commands;
  gmap_t *tmp,*tmp2;

  crop(src, &R, 0xff);
  reduce_colors(src,16);

  command_buffer= buffer_new(1024);
  if (!bitmap_tree)  bitmap_tree= objmap_new();
  if (!graymap_tree) graymap_tree= objmap_new();

  ncommands= 0;

  tmp= mkgtemp(src);
  tmp2= mkgtemp(src);

  remove_lines(src, command_buffer, &ncommands);
  remove_trivial_bitmaps(src, tmp, command_buffer, &ncommands);



  extract_bitmaps_with_background(src, tmp, tmp2, command_buffer, &ncommands);
  encode_images(src, command_buffer, &ncommands);
  free_gmap(tmp);
  free_gmap(tmp2);

  buffer_free(command_buffer, (void**) &commands, NULL);
  *Rncommands= ncommands;
  return commands;
}


/***********************************************************************
 *                                                                     *
 *                               Utility                               *
 *                                                                     *
 ***********************************************************************/

static void set_fpos(FILE *f, unsigned int pos)
{
  if (fseek(f, pos, SEEK_SET))
    exit(fprintf(stderr,"can't seek in output file: %s\n",strerror(errno)));
}


static unsigned int get_fpos(FILE *f)
{
  long L;
  L= ftell(f);
  if (L<0)
    exit(fprintf(stderr,"can't get file position for output: %s\n",
            strerror(errno)));
  return L;
}

static void seek_end(FILE *f)
{
  if (fseek(f, 0, SEEK_END))
    exit(fprintf(stderr,"can't seek in output file: %s\n",strerror(errno)));
}

static void write_bytes(FILE *f, void *data, int len)
{
  int R;
  R= fwrite(data, 1, len, f);
  if (R!=len)
    exit(fprintf(stderr,"error writing to output\n"));
}

static int write_padding(FILE *f, int sz)
{
  static char zeros[1000];
  int O= sz;
  while(sz)
  {
    int cp;
    cp= sizeof(zeros);
    if (cp>sz) cp= sz;
    write_bytes(f,zeros, cp);
    sz-= cp;
  }
  return O;
}

/***********************************************************************
 *                                                                     *
 *                               Output                                *
 *                                                                     *
 ***********************************************************************/


void shift(int *argc,char **argv,int i, int a)
{
  for(;i+a<*argc;i++)
    argv[i]= argv[i+a];
  *argc-= a;
}





static int write_bitmap_data(FILE *out, bentry_t *be)
{
  unsigned char TB[10];
  // bitmap on-disk format:
  // size[4] , data
  // conveniently, size is already in be->len.
  Odoub(TB, be->len);
  write_bytes(out, TB, 4);
  write_bytes(out, be->data, be->len);
  return be->len + 4;
}

static int write_graymap_data(FILE *out, gentry_t *ge)
{
  unsigned char TB[10];
  // bitmap on-disk format:
  // size[4] , data
  // conveniently, size is already in ge->len.
  Odoub(TB, ge->len);
  write_bytes(out, TB, 4);
  write_bytes(out, ge->data, ge->len);
  return ge->len + 4;
}

static void write_bitmap_stuff(FILE *out)
{
  int off;
  int i;
  unsigned char *table;
  unsigned char TB[40];

  unsigned int table_size, data_size, section_offset;
  
  table_size= 4*(bitmap_count+1);

  table= malloc(table_size);
  Odoub(table, bitmap_count);

  bitmap_array= malloc(sizeof(bitmap_array[0])*bitmap_count);
  for(i=0;i<bitmap_count;i++)
  {
    bitmap_array[i]= objmap_destroy(bitmap_tree);
    bitmap_array[i]->number= i;
  }

  off= get_fpos(out);

  section_offset= off;
  data_size= 0;

  off+= write_padding(out, table_size);

  for(i=0;i<bitmap_count;i++)
  {
    int R;
    Odoub(table + 4*(i+1), off);
    R= write_bitmap_data(out, bitmap_array[i]);
    off+= R;
    data_size+= R;
    free(bitmap_array[i]->data);
  }

  set_fpos(out, section_offset);
  write_bytes(out, table, table_size);

  set_fpos(out, HDR_BMAP);
  Odoub(TB, data_size + table_size);
  Odoub(TB+4, section_offset);
  write_bytes(out, TB, 8);

  seek_end(out);
  free(table);

printf("bitmap section size  = %d\n",data_size+table_size);
printf("bitmap section offset= %d\n", section_offset);
printf("number of bitmaps    = %d\n", bitmap_count);
}

/***    
      graymaps are essentially the same thing as bitmaps.
      I just used different types for them in order to
      not mess things up.                                       ***/

static void write_graymap_stuff(FILE *out)
{
  int off;
  int i;
  unsigned char *table;
  unsigned char TB[40];

  unsigned int table_size, data_size, section_offset;
  
  table_size= 4*(graymap_count+1);

  table= malloc(table_size);
  Odoub(table, graymap_count);

  graymap_array= malloc(sizeof(graymap_array[0])*graymap_count);
  for(i=0;i<graymap_count;i++)
  {
    graymap_array[i]= objmap_destroy(graymap_tree);
    graymap_array[i]->number= i;
  }

  off= get_fpos(out);

  section_offset= off;
  data_size= 0;

  off+= write_padding(out, table_size);

  for(i=0;i<graymap_count;i++)
  {
    int R;
    Odoub(table + 4*(i+1), off);
    R= write_graymap_data(out, graymap_array[i]);
    off+= R;
    data_size+= R;
    free(graymap_array[i]->data);
  }

  set_fpos(out, section_offset);
  write_bytes(out, table, table_size);

  set_fpos(out, HDR_GMAP);
  Odoub(TB, data_size + table_size);
  Odoub(TB+4, section_offset);
  write_bytes(out, TB, 8);

  seek_end(out);
  free(table);

printf("graymap section size  = %d\n",data_size+table_size);
printf("graymap section offset= %d\n", section_offset);
printf("number of graymaps    = %d\n", graymap_count);
}

/***
             bit: 'B' X[2] Y[2] B[4]
             gray: 'G' X[2] Y[2] G[4]
             rect: 'R' X[2] Y[2] W[2] H[2] color[1]          ***/

static unsigned char *serialize_commands(command_t *cmd, int ncmd,int *Rlen)
{
  unsigned char *buf;
  int sz,i,off;
  sz= 0;
  for(i=0;i<ncmd;i++)
  switch(cmd[i].type)
  {
  case 'G': sz+= 1+ 2+ 2+ 4; break;
  case 'B': sz+= 1+ 2+ 2+ 4; break;
  case 'R': sz+= 1+ 2+ 2+ 2+ 2+ 1; break;
  }

  buf= malloc(sz);
  *Rlen= sz;
  off= 0;
  // commands need to be written in the reverse order we saw them
  // because of reasons
  for(i=ncmd-1;i>=0;i--)
  switch(cmd[i].type)
  {
  case 'G': 
    off+= Obyte(buf+off, 'G');
    off+= Oword(buf+off, cmd[i].x);
    off+= Oword(buf+off, cmd[i].y);
    off+= Odoub(buf+off, cmd[i].d.G.ge->number);
    break;
  case 'B': 
    off+= Obyte(buf+off, 'B');
    off+= Oword(buf+off, cmd[i].x);
    off+= Oword(buf+off, cmd[i].y);
    off+= Odoub(buf+off, cmd[i].d.B.be->number);
    break;
  case 'R': 
    off+= Obyte(buf+off, 'R');
    off+= Oword(buf+off, cmd[i].x);
    off+= Oword(buf+off, cmd[i].y);
    off+= Oword(buf+off, cmd[i].d.R.w);
    off+= Oword(buf+off, cmd[i].d.R.h);
    off+= Obyte(buf+off, cmd[i].d.R.color);
    break;
  }
  return buf;
}


static void write_page_stuff(FILE *out)
{
  unsigned int table_size, data_size, section_offset, off;
  unsigned char *table, TB[50];
  int i,j;

  off= section_offset= get_fpos(out);
  table_size= 4*(1+page_count);
  data_size= 0;

  table= malloc(table_size);
  Odoub(table, page_count);

  off+= write_padding(out,table_size);

  ptrbuf_collect(page_buffer,(void***) &page_array, NULL);
  for(i=0;i<page_count;i++)
  {
    unsigned char *ps, *cs; int pl; int cl;
    ps= serialize_commands(page_array[i]->commands,
                           page_array[i]->ncommands,&pl);
    free(page_array[i]->commands);

    // this sets cl=pl and then pl=0 if the compression
    // fails or doesn't make sense. otherwise pl isn't touched
    // and the uncompressed buffer is freed.
    cs= optional_compress_commands(ps, &pl, &cl);
    if (cs!=ps) free(ps);
    ps= cs;

    Odoub(table+4*(i+1), off);
    j= 0;
                   // uncompressed command length, width, height, color
    j+= Odoub(TB+j, 4+ 2 + 2 + 1 + cl);         // size
    j+= Odoub(TB+j, pl);
    j+= Oword(TB+j, page_array[i]->w);
    j+= Oword(TB+j, page_array[i]->h);
    j+= Obyte(TB+j, 0xff);                  // color

    write_bytes(out, TB, j);
    write_bytes(out, ps, cl);
    off+= j+cl;
    data_size+= j+cl;
    free(ps);
  }

  set_fpos(out, section_offset);
  write_bytes(out, table, table_size);

  set_fpos(out, HDR_PAGE);  

  Odoub(TB, data_size+table_size);
  Odoub(TB+4, section_offset);
  write_bytes(out, TB, 8);

  free(table);
  seek_end(out);

printf("page section size  = %d\n", data_size+table_size);
printf("page section offset= %d\n", section_offset);
printf("number of pages=     %d\n", page_count);
}

static void process_page(gmap_t *gm)
{
  page_t *P;

  P= malloc(sizeof(*P));
  ptrbuf_put(page_buffer, P); page_count++;
  P->commands= process_image(gm, &P->ncommands);
  printf("page %4d : %4d x %4d\n", page_count,gm->w,gm->h);
  P->w= gm->w;
  P->h= gm->h;
  free(gm->data);
}

static void process_file(char *fn)
{
  gmap_t gm;
  char *ie;
  ie= read_png_graymap_noalpha(fn,(unsigned int*)&gm.w,
                                  (unsigned int*)&gm.h, &gm.data);
  if (ie)
     exit(fprintf(stderr,"can't read %s: %s\n", fn,ie));

  process_page(&gm);
}

static void process_stdin()
{
  gmap_t gm;
  char *ie;
  while(!feof(stdin))
  {
    int eof;

    ie= read_png_g8_filep(stdin,(unsigned int*)&gm.w,
                                (unsigned int*)&gm.h, &gm.data, &eof);
    if (ie)
    { 
       if (!eof) exit(fprintf(stderr,"can't read stdin: %s\n", ie));
       break;
    }
    process_page(&gm);
  }
}

static void write_bookmark_stuff(FILE *out,bmk_t *list)
{
  unsigned int section_offset;
  unsigned char *data; unsigned int data_size;
  unsigned char TB[20];

  section_offset= get_fpos(out);

  data= bmk_encode(list, &data_size);
  write_bytes(out, data, data_size);

  set_fpos(out, HDR_BMK);  

  Odoub(TB, data_size);
  Odoub(TB+4, section_offset);
  write_bytes(out, TB, 8);

  seek_end(out);
  free(data);

printf("bookmark section size  = %d\n", data_size);
printf("bookmark section offset= %d\n", section_offset);
}

int main(int argc,char **argv)
{
  char *ofile= "out.imd";
  FILE *out;
  char *bmk_file;
  int i;
  bmk_t *bmk;

  page_buffer= ptrbuf_new(50);
  page_count= 0;
  bmk_file= NULL;

  for(i=1;i<argc;i++)
   if (!strcmp(argv[i],"-o"))
   {
      int a;
      if (i<argc-1) { ofile= argv[i+1]; a=2; }
      else a= 1;

      shift(&argc,argv,i,a);
      i--; 
   }
   else if (!strcmp(argv[i],"-b"))
   {
      if (i<argc-1) { bmk_file= argv[i+1]; shift(&argc,argv,i,2); }
      else shift(&argc,argv,i,1);
      i--;
   }

  out= fopen(ofile,"wb");
  if (!out)
    exit(fprintf(stderr,"can't open %s for writing: %s\n", 
               ofile, strerror(errno)));

  if (bmk_file) bmk= bmk_parse(bmk_file); else bmk= NULL;
 
  for(i=1;i<argc;i++)
    if (!strcmp(argv[i],"-")) process_stdin();
                         else process_file(argv[i]);

  write_padding(out, HDR_TOTAL);

  write_bitmap_stuff(out);
  write_graymap_stuff(out);
  write_page_stuff(out);
  if (bmk) write_bookmark_stuff(out,bmk);

  printf("graymap hits = %d\n",graymap_hits);
  printf("bitmap hits  = %d\n",bitmap_hits);

  set_fpos(out,0);
  write_bytes(out, MAGIC, MAGIC_SIZE);
  fclose(out);
  return 0;
}
