#include <sys/stat.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>

#define MAX_LINE_LENGTH 256
#define BUFFER_SIZE     8192

struct iofil
{
  int fd;
  char *path;
};


void sdie(const char *fmt, ...)
{
  int eno;
  eno= errno;
  va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  fprintf(stderr,": %s\n", strerror(eno));
  exit(1);
}

void die(const char *fmt, ...)
{
  va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  fprintf(stderr,"\n");
  exit(1);
}

struct iofil* open_file(char *path, int mode)
{
  struct iofil *F;
  F= malloc(sizeof(*F));
  if (!F) sdie("malloc(%d)",(int)sizeof(*F));
  F->fd= open(path, mode, 0600);
  if (F->fd<0) sdie("open(%s)",path); 
  F->path= path;
  return F;
}

void close_file(struct iofil *F)
{
  if (close(F->fd)) sdie("close(%s)",F->path);
  free(F);
}

size_t read_bytes(struct iofil *F, char *dst, size_t len)
{
  ssize_t R;
  size_t T;
  T= 0;
  while(len)
  {
    R= read(F->fd, dst,len);
    if (R<0)
    {
      if (errno==EINTR) continue;
      sdie("read(%s)", F->path);
    }
    if (R==0) break;
    T+= R;
    dst += R;
    len -= R;
  }
  return T;
}

void write_bytes(struct iofil *F, char *src, size_t len)
{
  ssize_t R;
  while(len)
  {
    R= write(F->fd, src,len);
    if (R<0)
    {
      if (errno==EINTR) continue;
      sdie("write(%s)", F->path);
    }
    if (R==0) 
      die("write(%s,%d): unexpected return value 0", F->path, len);
    src += R;
    len -= R;
  }
}

ssize_t find_line(char *str, int skip, size_t len)
{
  char *T;
  T= memchr(str+skip, '\n', len - skip);
  if (!T) return -1;
  return T-str+1;
}

void write_spaces(struct iofil* ofil, int len)
{
  static char spaces[MAX_LINE_LENGTH];
  if (len>MAX_LINE_LENGTH)
     die("line length %d is too big (>%d)", len, MAX_LINE_LENGTH);
  if (len<=0)
     die("line length %d is invalid", len);
  memset(spaces, ' ', len-1);
  spaces[len-1]= '\n';
  write_bytes(ofil, spaces, len);
}

int ispc(char x) 
{ 
  if (x==127) return 1;
  if (x>=0 && x<=' ') return 1;
  return 0;
}

int match(char *s, int slen, char *pfx, char *sfx)
{
  char *e;
  int n;
  e= s+ slen;

  while(e>s && ispc(e[-1])) e--;
  if (e==s) return 0; // empty line of spaces

  while(s<e && ispc(s[0])) s++;
  if (s==e) return 0; // impossible, but anyway.

  if (pfx && pfx[0])
  {
    n= strlen(pfx);
    if (e-s<n) return 0;
    if (memcmp(s,pfx, n)) return 0;
  }

  if (sfx && sfx[0])
  {
    n= strlen(sfx);
    if (e-s<n) return 0;
    if (memcmp(e-n, sfx, n)) return 0;
  }
  return 1;
}


void process(struct iofil *ofil, struct iofil *ifil,int thr,
             char *pfx, char *sfx)
{
  char buf[BUFFER_SIZE];
  size_t buflen, bufstart;
  int le, skip;
  ssize_t R;

  buflen= 0;
  bufstart= 0;
  skip= 0;

restart:
  if (buflen==0)
  {
    bufstart= 0;
    buflen= read_bytes(ifil, buf, BUFFER_SIZE);
    if (!buflen) return ;
  }

again:
  le= find_line(buf+bufstart, skip, buflen);
  skip= 0;
  if (le==-1) // no line found
  {
    if (buflen>=thr)  
    {
      write_bytes(ofil, buf+bufstart, buflen);
      bufstart= buflen= 0;
      goto copy_to_next_line;
    }
    if (bufstart)
    {
      memmove(buf, buf+bufstart, buflen);
      bufstart= 0;
    }
    R= read_bytes(ifil, buf+buflen, BUFFER_SIZE-buflen);
    if (!R)
    {
      write_bytes(ofil, buf, buflen);
      return ; // no line and we're at the end of the file
    }
    skip= buflen;
    buflen += R;
    goto again;
  } 
  if (le<thr && match(buf+bufstart, le, pfx, sfx))
    write_spaces(ofil, le);
  else
    write_bytes(ofil, buf+bufstart, le);
  bufstart += le;
  buflen-= le;
  goto restart;
copy_to_next_line:
  // when we come here, bufstart= buflen= 0.
  buflen= read_bytes(ifil, buf, BUFFER_SIZE);
  if (buflen==0) return ; // all done
  le= find_line(buf, 0, buflen);
  if (le==-1)
  {
    write_bytes(ofil, buf, buflen);
    buflen= 0;
    goto copy_to_next_line;
  }
  write_bytes(ofil, buf, le);
  bufstart += le;
  buflen-= le;
  goto restart;
}

int main(int argc,char **argv)
{
  char *ifn, *ofn;
  int thr;
  int i;
  char *pfx, *sfx;
  struct iofil *ifil, *ofil;

  if (argc!=6)
    die("usage: %s input-fn output-fn threshold prefix suffix", argv[0]);
  
  i= 1;
  ifn= argv[i++];
  ofn= argv[i++];
  thr= atoi(argv[i++]);
  if (thr<=0) die("bad threshold %d", thr);
  pfx= argv[i++];
  sfx= argv[i++];
  if (!pfx[0] &&sfx[0]) die("prefix and suffix can't be both empty");

  ifil= open_file(ifn, O_RDONLY);
  ofil= open_file(ofn, O_WRONLY | O_CREAT | O_TRUNC);
  process(ofil, ifil,thr, pfx, sfx);
  close_file(ifil);
  close_file(ofil);
  return 0;
}
